mod paths;
mod encoder;
mod storage;
mod editor;
mod git;


use storage::{Storage, Item};

use clap::{Parser, Subcommand};
use rpassword;

use std::io::{self, Write};
use std::process;


#[derive(Parser)]
#[command(name = "mps", version = "0.0.1", about = "MyPasswordStorage: Tool for storing your passwords locally with git synchronization")]
struct Cli {
    #[command(subcommand)]
    command: Option<Commands>,
}

#[derive(Subcommand)]
enum Commands {

    /// Initialization of storage and config, use this in first time of usage of mps
    Init,

    /// Lists all ids stored in db
    List,

    /// Show content of an item
    Show {
        #[arg(value_name="item_id")]
        id: String
    },
    
    /// Adds new item with unique id to the storage
    Add {
        #[arg(value_name="item_id")]
        id: String
    },

    /// Edit item content
    Edit {
        #[arg(value_name="item_id")]
        id: String
    },

    Delete {
        #[arg(value_name="item_id")]
        id: String
    }

}

enum PROMPT {
    YES,
    NO
}

fn get_prompt(question: &str) -> io::Result<PROMPT> {
    print!("{} [Y/n] ", question);
    io::stdout().flush()?;
    let mut input = String::new();
    io::stdin().read_line(&mut input)?;
    let input = input.trim().to_lowercase();
    match input.as_str() {
        "y" => Ok(PROMPT::YES),
        "yes" => Ok(PROMPT::YES),
        "n" => Ok(PROMPT::NO),
        "no" => Ok(PROMPT::NO),
        _ => Ok(PROMPT::YES),
    }

}

struct MPS {
    storage: Option<Storage>
}

impl MPS {
    pub fn new() -> MPS {
        MPS { storage: None }
    }

    fn print_init_hint() {
        println!("1. Create empty repository for storing your passwords on gitlab or somewhere you prefer");
        println!("2. Create dir `mps`: `mkdir ~/mps`");
        println!("3. Clone to your home folder `~/.mps/storage`: `git clone <your_storage_url> storage`");
        println!("4. Export varable `$MPS_HOME`: `export MPS_HOME=\"/home/<your_username>/.mps\"`");
        println!("5. Run `mps init`");
    }
    
    fn prompt_new_password() -> io::Result<String> {
        print!("Enter passphrase: ");
        io::stdout().flush()?;
        let ps  = rpassword::read_password()?;
        if ps.len() < 8 {
            return Err(io::Error::new(io::ErrorKind::InvalidInput, "Passphrase must be longer that 8 letters"));
        }
        if ps.len() > 32 {
            return Err(io::Error::new(io::ErrorKind::InvalidInput, "Passphrase must not be longer than 32 letters"));
        }
        print!("Enter same passphrase again: ");
        io::stdout().flush()?;
        let ps2 = rpassword::read_password()?;
        if ps != ps2 {
            return Err(io::Error::new(io::ErrorKind::InvalidInput, "Passphrases must be equal"));
        }
        Ok(ps)
    }

    fn init() -> io::Result<()> {
        if Storage::is_inited()? {
            return Err(io::Error::new(io::ErrorKind::AlreadyExists, "Reinitialization attempted"));
        }
        let passphrase;
        loop {
            match MPS::prompt_new_password() {
                Ok(ps) => {
                    passphrase = ps;
                    break;
                },
                Err(e) => {
                    println!("{}", e);
                    continue
                }
            }
        }
        Storage::init(passphrase)?;
        Ok(())
    }

    pub fn login(&mut self) -> io::Result<()> {
        if self.storage.is_some() {
            return Ok(());
        }
        print!("Enter passphrase for storage: ");
        io::stdout().flush()?;
        let passphrase = rpassword::read_password()?;
        print!("\x1B[1A\x1B[2K");  // Move cursor up one line and clear that line
        io::stdout().flush()?;
        // storage will return error if passphrase is incorrect
        self.storage = Some(Storage::from_db(passphrase)?);
        Ok(())
    }

    fn list(&mut self) -> io:: Result<()> {
        self.login()?;
        let ids = self.storage.as_ref().unwrap().ids();
        // let mut counter = 1;
        if ids.len() == 0 {
            println!("No items");
        } else {
            for id in ids {
                println!("{}", id);
                //counter += 1;
            }
        }
        Ok(())
    }

    fn show(&mut self, id: &String) -> io::Result<()> {
        self.login()?;
        let st = self.storage.as_ref().unwrap();
        if !st.contains(id) {
            return Err(io::Error::new(
                io::ErrorKind::InvalidInput,
                format!("Can`t find id: {}", id)
            ));
        }
        let item = st.get(id);
        editor::open_to_show(&item.content)?;
        Ok(())
    }

    fn add(&mut self, id: &String) -> io::Result<()> {
        self.login()?;
        let st = self.storage.as_mut().unwrap();
        if st.contains(id) {
            let question = format!("Item [{}] exist. Do you want to edit it instead?", id);
            match get_prompt(&question)? {
                PROMPT::YES => {
                    self.edit(id)?;
                    return Ok(());
                },
                PROMPT::NO => return Ok(()),
            }
        }
        // set empty string because there is no content yet
        let content = editor::open_to_edit(&String::from(""))?;
        st.add(Item::from(id.clone(), content));
        st.dump()?;
        Ok(())
    }

    fn edit(&mut self, id: &String) -> io::Result<()> {
        self.login()?;
        let st = self.storage.as_mut().unwrap();
        if !st.contains(id) {
            let question = format!("Item [{}] exist. Do you want to add it instead?", id);
            match get_prompt(&question)? {
                PROMPT::YES => {
                    self.add(id)?;
                    return Ok(());
                }
                PROMPT::NO => return Ok(())
            }
        }
        let mut item = (*st.get(id)).clone();
        let new_content = editor::open_to_edit(&item.content)?;
        item.content = new_content;
        st.update(item);
        st.dump()?;
        Ok(())
    }

    fn delete(&mut self, id: &String) -> io::Result<()> {
        self.login()?;
        let st = self.storage.as_mut().unwrap();
        if !st.contains(id) {
            return Err(io::Error::new(
                io::ErrorKind::InvalidInput,
                "No such item"
            ));
        }
        st.remove(id);
        st.dump()?;
        Ok(())
    }

}

fn run_command() -> io::Result<()> {
    let cli = Cli::parse();
    match &cli.command {
        Some(Commands::Init) => {
            MPS::init()?;
        }
        Some(Commands::List) => {
            let mut mps = MPS::new();
            mps.list()?;
        }
        Some(Commands::Show { id }) => {
            let mut mps = MPS::new();
            mps.show(id)?;
        }
        Some(Commands::Add { id }) => {
            let mut mps = MPS::new();
            mps.add(id)?;
        }
        Some(Commands::Edit { id }) => {
            let mut mps = MPS::new();
            mps.edit(id)?
        }
        Some(Commands::Delete { id }) => {
            let mut mps = MPS::new();
            mps.delete(id)?
        }
        None => {
            match Storage::check_installed() {
                Ok(()) => (),
                Err(e) => {
                    println!("{}", e);
                    MPS::print_init_hint();
                    return Ok(());
                }
            }
            if !Storage::is_inited()? {
                match get_prompt("Do you want to init storage db?")? {
                    PROMPT::YES => MPS::init()?,
                    PROMPT::NO => return Ok(()),
                }
            } else {
                let mut mps = MPS::new();
                mps.list()?
            }
        }
    }
    Ok(())
}

fn main() {
    match run_command() {
        Ok(()) => return,
        Err(e) => {
            println!("{}", e);
            process::exit(2); // TODO: better codes for different errors
        }
    }
}

